/* Coincidence Timer "PhotoFinish".  Started Dec 2007 by Alan Stummer astummer_at_physics.utoronto.ca

	RAISON D'ETRE
	PhotoFinish measures the time difference between photodiode, from about 100pS to 150nS.
	
	MODUS OPERANDI
	Two ADCs sample at the 100MHz clock rate.  When a photodiode event happens on either channel,
	the hardware trips and ramps an RC integrator.  The ADC asynchronously (to the event) measures
	both RC integrators at 10nS intervals.  If either ADC level is above threshold then the 
	event cycle is triggered.  If both ADCs are above threshold before timeout (about 1uS) 
	then a coincidence has occured.  A flag is set, the Rabbit processor reads the ADCs and the 
	number of cycles between the photodide events, then acknowledges the coicidence and the cycle 
	starts again.  By knowing the RC "1" and "0" levels and the RC time constant, the photodiode 
	event times can be calculated and therefore the time difference between them is known.  Measured
	resolution is <100pS.
	DATA TRANSFER HANDSHAKING: When a coincidence is detected, DataReady is put high to signal the
	Rabbit processor. Rabbit reads the data then acknowledges by lowering RunEn[able]. DataReady is 
	lowered and the state machine restarts. Coincidence detection does not begin until Rabbit sets
	RunEn[able] high.

	See file "PhotoFinish Protocols.txt" for details of registers, commands and addressing.
*/

// Generated by Quartus II Version 7.2 (Build Build 151 09/26/2007)
// Created on Wed Dec 19 08:54:34 2007

//synthesis VERILOG_INPUT_VERSION SYSTEMVERILOG_2005

module uberPhotoFinish (
	input      [11:0] Da, Db,									  //the two raw ADC buses, in 2's complement.
	input             OrA, OrB,															 //over range from ADCs.
	input             Clk,																	//100MHz clock from PLL.
	input      [2:0]  Arabbit,															 //address bus from Rabbit.
	input             RunEnIn,						 //Run enable in and Rabbit acknowledging coincidence.
	input             CalIn, CalHiIn,											 //calibrate enable and level.
	output reg [11:0] Drabbit,																	//data bus to Rabbit.
	output reg        DataReady = 0, 					  //flag out, coincidence detected and measured.
	output            CalOut, CalHiOut );										 //calibrate enable and level.

	reg [7:0]  CycleTimer;								 //number of clocks after either channel is active.
	reg [11:0] Reg [2:0];										 //ADC values and offset after a coincidence.
	reg [11:0] AdcA, AdcB;								//ADC values in plain old unsigned straight binary.
	reg [2:0]  State = 3'd0;																//state machine pointer.
	reg [15:0] SinglesA, SinglesB;										 //single hits per channel counters.
	reg [2:0]  ArabbitBuf;							  //double buffer for asynchronous Rabbit address bus.
	reg [2:0]  LastArabbit;	//last Rabbit address, because data bus changes on address changes only.
	reg [11:0] Da1, Db1;																			 //pipelined buffers.
	reg        RunEn1, RunEn;																	 //pipelined buffers.
	reg [11:0] THRESHOLD   = 12'd525;		//trigger threshold, to detect when a channel is tripped.

	parameter VERSION     = 12'h110;				  //version: 4-bit major, 4-bit minor, 4-bit revision.
//	parameter THRESHOLD   = 12'd525;			//trigger threshold, to detect when a channel is tripped.
	parameter TIMEOUT     = 8'd100;		//number of 10nS clocks before assuming single channel only.
	parameter RESET_TIME  = 8'd100;					//number of 10nS cycles to allow channels to reset.

	//Note: to convert ADC data in 2's complement to flat coding use ADC = (ADC + 2048) modulus 4096.
	//The modulus is inherent in the word size so only have to add 2^11.


/************************************ Present some data to Rabbit ********************************/
	always @ (posedge Clk)  begin
		Da1  <= Da;
		AdcA <= Da1 + 12'd2048;											  //ADC A in plain old unsigned binary.
		Db1  <= Db;
		AdcB <= Db1 + 12'd2048;											  //ADC B in plain old unsigned binary.
		CalOut   <= CalIn;							  //pipe calibration controls through, for future use.
		CalHiOut <= CalHiIn;
		RunEn1 <= RunEnIn;																	 //Run Enable buffering.
		RunEn  <= RunEn1;	
		ArabbitBuf <= Arabbit;																	//buffer address bus.

		if (ArabbitBuf != LastArabbit)  begin		  //OUTPUT DATA TO THE RABBIT WHEN ADDRESS CHANGES.
			LastArabbit <= ArabbitBuf;
			case (ArabbitBuf)			
				3'd0:  Drabbit <= Reg[0];													 //ADC A at coincidence.
				3'd1:  Drabbit <= Reg[1];													 //ADC B at coincidence.
				3'd2:  Drabbit <= Reg[2];													//offset and overranges.
				3'd3:  Drabbit <= AdcA;										 //ADC A in binary, for calibration.
				3'd4:  Drabbit <= AdcB;										 //ADC B in binary, for calibration.
				3'd5:  Drabbit <= SinglesA[11:0];										  //Singles A, LS chunk.
				3'd6:  Drabbit <= SinglesB[11:0];										  //Singles B, LS chunk.
				3'd7:  begin																	  //Singles, MS nibbles.
							Drabbit[3:0]  <= SinglesA[15:12];
							Drabbit[7:4]  <= 4'd0;
							Drabbit[11:8] <= SinglesB[15:12];
						 end
			endcase
		end
	end


/************************************** Main State Machine ***************************************/
	always @ (negedge Clk)  begin
		case (State)
			3'd0:	 begin																 //WAIT FOR RUN TO BE ENABLED.
				DataReady  <= 0;
				CycleTimer <= 8'd0;
				if (!CalIn & CalHiIn)  begin											 //reset singles counters.
					SinglesA <= 12'd0;
					SinglesB <= 12'd0;
				end
				if (RunEn)  State <= 3'd1;
			end


			3'd1: begin										  //WAIT FOR BOTH ADCS TO BE LOW BEFORE CONTINUING.
				if ((AdcA < THRESHOLD) & (AdcB < THRESHOLD))  State++;
			end
	
 
			3'd2: begin						  //ALLOW TIME FOR BOTH ADCS TO BE FULLY LOW BEFORE CONTINUING.
					if (CycleTimer++ > RESET_TIME)  State++;
			end
	
 
			3'd3:  begin									 					 //WAIT FOR CHANNEL(S) TO BE ACTIVE.
				if (!RunEn)  State <= 3'd0;												//run no longer enabled.
				CycleTimer <= 8'd0;
				if ((AdcA > THRESHOLD) & (AdcB < THRESHOLD))	begin						  //Only Ch A Active.
					Reg[0]      <= AdcA;																	  //save ADC A.
					Reg[2][8]   <= OrA;																	 //overrange A.
					Reg[2][10]  <= 1;																	//A tripped first.
					if(SinglesA < 16'hFFFF)  SinglesA <= SinglesA + 16'd1;
					State       <= 3'd4;
				end
	
				else if ((AdcA < THRESHOLD) & (AdcB > THRESHOLD))	begin				  //Only Ch B Active.
					Reg[1]      <= AdcB;																	  //save ADC B.
					Reg[2][9]   <= OrB;																	 //overrange B.
					Reg[2][10]  <= 0;																	//B tripped first.
					if(SinglesB < 16'hFFFF)  SinglesB <= SinglesB + 16'd1;
					State       <= 3'd5;
				end

				else  if ((AdcA > THRESHOLD) & (AdcB > THRESHOLD))	begin						 //Both Active.
					Reg[0]      <= AdcA;																		//save ADCs.
					Reg[1]      <= AdcB;
					Reg[2][7:0] <= 8'd0;																	 //zero offset.
					Reg[2][8]   <= OrA;														 //save overrange flags.
					Reg[2][9]   <= OrB;
					if (AdcA > AdcB)  Reg[2][10]  <= 1;  else  Reg[2][10]  <= 0;			  //A first?
					if(SinglesA < 16'hFFFF)  SinglesA <= SinglesA + 16'd1;
					if(SinglesB < 16'hFFFF)  SinglesB <= SinglesB + 16'd1;
					State       <= 3'd6;
				end 
			end


			3'd4: begin																	 //CH A ACTIVE, WAIT FOR CH B.
				if (AdcB > THRESHOLD)  begin															 //Coincidence!
					Reg[1]      <= AdcB;																	  //save ADC B.
					Reg[2][9]   <= OrB;
					Reg[2][7:0] <= ++CycleTimer;														 //save offset.
					if(SinglesB < 16'hFFFF)  SinglesB <= SinglesB + 16'd1;
					State       <= 3'd6;
				end
				else if (CycleTimer++ > TIMEOUT)  State <= 3'd0;				//timeout, no channel B.
			end


			3'd5: begin																	 //CH B ACTIVE, WAIT FOR CH A.
				if (AdcA > THRESHOLD)  begin															 //Coincidence!
					Reg[0]      <= AdcA;																	  //save ADC A.
					Reg[2][8]   <= OrA;
					Reg[2][7:0] <= ++CycleTimer;														 //save offset.
					if(SinglesA < 16'hFFFF)  SinglesA <= SinglesA + 16'd1;
					State       <= 3'd6;
				end
				else if (CycleTimer++ > TIMEOUT)  State <= 3'd0;				//timeout, no channel B.
			end


			3'd6: begin										//HANDSHAKE WITH RABBIT TO ACKNOWLEDGE COINCIDENCE.
				if (!RunEn)  begin							//Rabbit has lowered RunEn as acknowledgment.
					DataReady  <= 0;		//drop flag asap before Rabbit thinks it's another coincidence.
					State      <= 3'd0;
				end
				else DataReady <= 1;						 //flag (for Rabbit) to say have valid coincidence.
			end


			default:  State <= 3'd0;
		endcase
	end

endmodule
