/* Coincidence Timer "PhotoFinish".  Started Dec 2007 by Alan Stummer astummer_at_physics.utoronto.ca
	
	STILL TO DO
	==========================
	* See why actual coin window extends to 80nS when TIMEOUT = 5.
	* Using a different acknowledgement signal other than RunEn, to avoid losing counts.

	RAISON D'ETRE
	PhotoFinish measures the time difference between photodiode, from about 100pS to 150nS.
	
	MODUS OPERANDI
	Two ADCs sample at the 100MHz clock rate.  When a photodiode event happens on either channel,
	the hardware trips and ramps an RC integrator.  The ADC asynchronously (to the event) measures
	both RC integrators at 10nS intervals.  If either ADC level is above threshold then the 
	event cycle is triggered.  If both ADCs are above threshold before timeout (about 1uS) 
	then a coincidence has occured.  A flag is set, the Rabbit processor reads the ADCs and the 
	number of cycles between the photodide events, then acknowledges the coicidence and the cycle 
	starts again.  By knowing the RC "1" and "0" levels and the RC time constant, the photodiode 
	event times can be calculated and therefore the time difference between them is known.  Measured
	resolution is <100pS.
	DATA TRANSFER HANDSHAKING: When a coincidence is detected, DataReady is put high to signal the
	Rabbit processor. Rabbit reads the data then acknowledges by lowering RunEn[able]. DataReady is 
	lowered and the state machine restarts. Coincidence detection does not begin until Rabbit sets
	RunEn[able] high.

	See file "PhotoFinish Protocols.txt" for details of registers, commands and addressing.
*/

// Generated by Quartus II Version 7.2 (Build Build 151 09/26/2007)
// Created on Wed Dec 19 08:54:34 2007

//synthesis VERILOG_INPUT_VERSION SYSTEMVERILOG_2005

module uberPhotoFinish (
	input      [11:0] Da, Db,							//the two raw ADC buses, in 2's complement.
	input             OrA, OrB,												//over range from ADCs.
	input             Clk,												   //100MHz clock from PLL.
	input      [3:0]  Arabbit,											 //address bus from Rabbit.
	input             RunEnIn,				  //Run enable in and Rabbit acknowledging coincidence.
	input             RabbitReadingData,			  //reading data, count but don't measure coin.
	input             CS,									 //CS from Rabbit, enables address bus.
	output reg [11:0] Drabbit,												  //data bus to Rabbit.
	output reg        DataReady = 0, 				 //flag out, coincidence detected and measured.
	output 		      CalOut, CalHiOut );			  				  //calibrate enable and level.

	localparam VERSION     = 12'h125;			 //Verilog version: Version, sub-version, revision.
	localparam THRESHOLD   = 12'd525;	  //trigger threshold, to detect when a channel is tripped.
	localparam TIMEOUT     = 8'd9;	   //number of 10nS clocks before assuming single channel only.
	localparam RESET_TIME  = 8'd3;				//number of 10nS cycles to allow channels to reset.
	localparam MAX_HITS = 24'hFFFFFF;	   		 	//maximum number of hits then freezes (16.7E6).

	reg [3:0]  CycleTimer;						 //number of clocks after either channel is active.
	reg [11:0] Reg [2:0];						       //ADC values and offset after a coincidence.
	reg [11:0] AdcA, AdcB;						//ADC values in plain old unsigned straight binary.
	reg [2:0]  State = 3'd0;											   //state machine pointer.
	reg [23:0] HitsA, HitsB;										   //hits per channel counters.
	reg [23:0] CoinAB;									 //Coincidence counter with no time stamps.
	reg [3:0]  ArabbitBuf;					   //double buffer for asynchronous Rabbit address bus.
	reg        CSbuf;							    	//double buffer for asynchronous Rabbit CS.
	reg [11:0] Da1, Db1;													   //pipelined buffers.
	reg        RunEn1, RunEn;												   //pipelined buffers.
	reg		   CalOutBuf  = 0;					//pipelined buffers, use this one for if condition.
	reg		   RbRdData, RbRdData1, RdFlag;							//buffer for RabbitReadingData.
	reg		   WaitRabbit = 0;									//register equivalent of DataReady.
	reg [26:0] WaitTime	  = 27'd0; 			   //measures how long the Rabbit acknowlegement takes.
												  //shift 3-bits in the end to have 80ns increment.
		
	//Note: to convert ADC data in 2's complement to flat coding use ADC = (ADC + 2048) modulus 4096.
	//The modulus is inherent in the word size so only have to add 2^11.


/************************************** Calibration and Run Mode Control *************************/
//	assign CalOut = !RunEn;


/************************************** Main State Machine to Detect Coincidences ****************/
	always @ (negedge Clk)  begin
		if (!RbRdData1 & RbRdData) begin		  		  //detecting the falling edge of RbRdData.
			DataReady  <= 0;
			WaitRabbit <= 0;
		end
		case (State)
			3'd0:	 begin											  //WAIT FOR RUN TO BE ENABLED.
				if (!WaitRabbit) Reg[2]     <= 12'd0;				 //just to fill Reg[2][7:4,11].
					
				CycleTimer <= 4'd0;					
				if (CSbuf & ArabbitBuf==4'd9)  begin					     //reset hits counters.
					HitsA   <= 24'd0;
					HitsB   <= 24'd0;
					CoinAB  <= 24'd0;
					WaitTime <= 27'd0;
				end
				else if (RbRdData) WaitTime <= WaitTime + 27'd1;
				if (RunEn)  State <= 3'd1;
			end


			3'd1: begin							  //WAIT FOR BOTH ADCS TO BE LOW BEFORE CONTINUING.
				if (!RunEn)  State <= 3'd0;				
				else if ((AdcA < THRESHOLD) & (AdcB < THRESHOLD))  State <= 3'd2;
			end
	
 
			3'd2: begin				  //ALLOW TIME FOR BOTH ADCS TO BE FULLY LOW BEFORE CONTINUING.
				CycleTimer <= CycleTimer + 4'd1;
				if (!RunEn)  State <= 3'd0;
				else if (CycleTimer > RESET_TIME)  State	<= 3'd3;
			end
	
 
			3'd3:  begin									 	//WAIT FOR CHANNEL(S) TO BE ACTIVE.
				CycleTimer <= 4'd1;											   //for state 4 and 5.
				if (!RunEn)  State <= 3'd0;								   //run no longer enabled.
				else	if ((AdcA > THRESHOLD) & (AdcB < THRESHOLD))	begin	//Only Ch A Active.
					if (!WaitRabbit) begin		  //refresh the arrival time only when not waiting.
						Reg[0]      <= AdcA;										  //save ADC A.
						Reg[2][8]   <= OrA;											 //overrange A.
						Reg[2][10]  <= 1;										 //A tripped first.
					end
					if (HitsA < MAX_HITS)  HitsA <= HitsA + 24'd1;					   //count hit.
					State       <= 3'd4;
				end
	
				else if ((AdcA < THRESHOLD) & (AdcB > THRESHOLD))	begin		//Only Ch B Active.
					if (!WaitRabbit) begin		  //refresh the arrival time only when not waiting.
						Reg[1]      <= AdcB;										  //save ADC B.
						Reg[2][9]   <= OrB;											 //overrange B.
						Reg[2][10]  <= 0;										 //B tripped first.
					end
					if (HitsB < MAX_HITS)  HitsB <= HitsB + 24'd1;					   //count hit.
					State       <= 3'd5;
				end

				else  if ((AdcA > THRESHOLD) & (AdcB > THRESHOLD))	begin			 //Both Active.
					if (!WaitRabbit) begin		  //refresh the arrival time only when not waiting.
						Reg[0]      <= AdcA;										   //save ADCs.
						Reg[1]      <= AdcB;
						Reg[2][3:0] <= 4'd0;										 //zero offset.
						Reg[2][8]   <= OrA;									//save overrange flags.
						Reg[2][9]   <= OrB;
						if (AdcA > AdcB)  Reg[2][10]  <= 1;  else  Reg[2][10]  <= 0;	 //A first?
					end
					if (HitsA < MAX_HITS)  HitsA  <= HitsA  + 24'd1;				   //count hit.
					if (HitsB < MAX_HITS)  HitsB  <= HitsB  + 24'd1;				   //count hit.
					if (CoinAB< MAX_HITS)  CoinAB <= CoinAB + 24'd1;		   //count coincidence.
					State       <= 3'd6;
				end 
			end


			3'd4: begin												  //CH A ACTIVE, WAIT FOR CH B.
				CycleTimer <= CycleTimer + 4'd1;
				if (!RunEn) State <= 3'd0;
				else if (AdcB > THRESHOLD)  begin									 //Coincidence!
					if (!WaitRabbit) begin		  //refresh the arrival time only when not waiting.
						Reg[1]      <= AdcB;										  //save ADC B.
						Reg[2][9]   <= OrB;
						Reg[2][3:0] <= CycleTimer;								 //save offset.
					end
					if (HitsB < MAX_HITS)  HitsB <= HitsB + 24'd1;					   //count hit.
					if (CoinAB< MAX_HITS)  CoinAB <= CoinAB + 24'd1;		   //count coincidence.
					State       <= 3'd6;
				end
				else if (CycleTimer == TIMEOUT)  State <= 3'd0;		   //timeout, no channel B.
			end


			3'd5: begin												  //CH B ACTIVE, WAIT FOR CH A.
				CycleTimer <= CycleTimer + 4'd1;
				if (!RunEn) State <= 3'd0;
				else if (AdcA > THRESHOLD)  begin									 //Coincidence!
					if (!WaitRabbit) begin		  //refresh the arrival time only when not waiting.
						Reg[0]      <= AdcA;										  //save ADC A.
						Reg[2][8]   <= OrA;
						Reg[2][3:0] <= CycleTimer;								 //save offset.
					end
					if (HitsA < MAX_HITS)  HitsA  <= HitsA  + 24'd1;				   //count hit.
					if (CoinAB< MAX_HITS)  CoinAB <= CoinAB + 24'd1;		   //count coincidence.
					State       <= 3'd6;
				end
				else if (CycleTimer == TIMEOUT)  State <= 3'd0;		   //timeout, no channel B.
			end


			3'd6: begin									  //FLAG TO RABBIT FOR COINCIDENCE ARRIVAL.
				if (RbRdData1 | !RbRdData) begin	  //set only when not falling edge of RbRdData.
					DataReady  <= 1;
					WaitRabbit <= 1;
				end
				State      <= 3'd0;
			end

			default:  State <= 3'd0;
		endcase
	end


/************************************ Present some data to Rabbit ********************************/
	always @ (posedge Clk)  begin
		Da1        <= Da;
		AdcA       <= Da1 + 12'd2048;						  //ADC A in plain old unsigned binary.
		Db1        <= Db;
		AdcB       <= Db1 + 12'd2048;						  //ADC B in plain old unsigned binary.
		RunEn1     <= RunEnIn;												//Run Enable buffering.
		RunEn      <= RunEn1;	
		ArabbitBuf <= Arabbit;												  //buffer address bus.
		CSbuf      <= CS;													  //buffer chip select.
		CalOut	   <= CalOutBuf;												   //CalOut buffer.
		RbRdData1  <= RabbitReadingData;
		RbRdData   <= RbRdData1;
				
		if (CSbuf)  begin												 //EXECUTE RABBIT COMMANDS.
			if ((ArabbitBuf != 4'd3)  & (ArabbitBuf != 4'd4))
				CalOutBuf <= (ArabbitBuf == 4'd10) | (ArabbitBuf == 4'd11);
			           
			case (ArabbitBuf)			
				4'd0:	Drabbit  <= Reg[0];								    //ADC A at coincidence.
				4'd1:	Drabbit  <= Reg[1];									//ADC B at coincidence.
				4'd2:	Drabbit  <= Reg[2];						   //offset, overranges, alignment.
				4'd3:	begin
					if (CalOutBuf)	Drabbit  <= AdcA;			//ADC A in binary, for calibration.
					else Drabbit <= WaitTime[14:3];
				end
				4'd4:	begin
					if (CalOutBuf)	Drabbit  <= AdcB;			//ADC B in binary, for calibration.
					else Drabbit <= WaitTime[26:15];
				end
				4'd5:	Drabbit  <= HitsA[11:0];						  		   //Hits A [11:0].
				4'd6:	Drabbit  <= HitsB[11:0];								   //Hits B [11:0].
				4'd7:	Drabbit  <= HitsA[23:12];								  //Hits A [23:12].
				4'd8:	Drabbit  <= HitsB[23:12];								  //Hits B [23:12].
				4'd9:	;										//reset hits counters (done above).
				4'd10:	CalHiOut  <= 0;											   //Calibrate low.
				4'd11: 	CalHiOut  <= 1;											  //Calibrate high.
  				4'd12:	;			   			   //not used. Was to take out of calibration mode.
				4'd13:  Drabbit <= VERSION;			   //Verilog version word and kill calibration.
				4'd14: 	Drabbit  <= CoinAB[11:0];					 //Coincidence AB count [11:0].
				4'd15:  Drabbit  <= CoinAB[23:12];					//Coincidence AB count [23:12].
				default: ;											 //unknown command, do nothing.
			endcase			
		end  //ends if CSbuf...
	end
endmodule
